<!-- omit in toc -->
# VS Code & CrossSection Module Demo
**Friday, August 21st, 2020**

Hello! 

The goal of this document is to provide an in-text version of all the things I am showing off at today's lab meeting. Namely, I would like to document the features I showcase in **VSCode** and in the Julia **CrossSection** module I created as part of the RadiativeTransfer package I am working on. 

Let's get started! 

<!-- omit in toc -->
## Table of Contents

- [VSCode](#vscode)
  - [Introduction](#introduction)
  - [Getting Started](#getting-started)
  - [Features List](#features-list)
- [Absorption](#absorption)
  - [Description](#description)
  - [Demo](#demo)

## VSCode 

### Introduction

When I first think of Visual Studio (the "VS" in VSCode), I immediately think of the clunky Microsoft IDE that I used many years ago to write C++ code. It was quite difficult to use and frankly, neither elegant nor approachable. 

In this document, I advocate for **Visual Studio Code**, which is confusingly, much different from **Visual Studio**. The former is a lightweight code editor like Sublime Text or Atom, whereas the latter is a full IDE like Eclipse or NetBeans. 

The primary reason I am vouching for VSCode, the code editor, is that it has completely improved my remote work experience. Previously, I was using `scp` or `git` pushing/pulling regularly to keep files in sync between my local system and remote servers (Fluo/Tofu). 

Now using VSCode, I can directly edit files on-server using a graphical code editor like Sublime. It allows you to perform large computations on Tofu and then directly view resulting plots inside the editor. The code editor is also language-agnostic, so there is great support for all mainstream coding languages. You can even run a Jupyter notebook inside VSCode. I can go on and on... Please see the features list [below](#features-list). 

### Getting Started

1. **Required:** You can download VSCode [here](https://code.visualstudio.com).
2. **Recommended:** Details on setting up the Remote SSH Extension to VSCode can be found [here](https://code.visualstudio.com/docs/remote/ssh). I would *highly* recommend this step to everyone! It has made working from home so much more pleasant for me. 
3. **Recommended:** You can set up language-specific extensions via the VSCode Extensions Marketplace (everything is free!). Linking the [Julia](https://marketplace.visualstudio.com/items?itemName=julialang.language-julia), [Python](https://marketplace.visualstudio.com/items?itemName=ms-python.python), [R](https://marketplace.visualstudio.com/items?itemName=Ikuyadeu.r), [Matlab](https://marketplace.visualstudio.com/items?itemName=Gimly81.matlab), and [other](https://marketplace.visualstudio.com/search?target=VSCode&category=Programming%20Languages&sortBy=Installs) languages.  
4. Optional: You can set up Jupyter Notebooks for Julia/Python/R inside VSCode. Although you can use any of these languages inside Jupyter after it is installed, Jupyter itself is installed only through the Python language extension found [here](https://marketplace.visualstudio.com/items?itemName=ms-python.python). After it is set up, you can switch the kernel/language in the top right corner of an open notebook. Fun fact: It's named Jupyter after the three languages it was built to support: JUlia, PYThon and R. 
5. Optional: There is a cool VSCode extension called [Live Share](https://marketplace.visualstudio.com/items?itemName=MS-vsliveshare.vsliveshare-pack) that allows multiple users to collaboratively edit code and debug in real-time (sort of like Google Docs but for editing and running code). I have not used this, but I have heard great things from the other Schmidt Scholars that this works really well. 

### Features List

I mentioned some of the main features in the [Getting Started](#Getting-Started) section above. Here is my personal list of VSCode's most useful features: 
- **Remote editing** files on Tofu/Fluo with a graphical text editor
- **Accessible terminal pane** open within the code editor to switch back and forth from the command line
- **IDE-like support**: I can't speak for other languages, but using the Julia extension, I have a full IDE with a convenient workspace to view all existing variables, their scope, documentation, etc. I believe the other language extensions provide similar functionality. 
- **Debugging**: Related to the previous point, but language-extensions enable debugging tools, which are supremely useful for setting breakpoints, seeing variable values in the middle of the code, etc. 
- **Jupyter notebooks**: Full support for Python/Julia notebooks running either remotely or locally, and editable directly inside the editor
- **Full-fledged git support**: Shortcuts for git commands, showing in-line version-history of code, simple interface for handling for merge-conflicts, etc. See more [here](https://code.visualstudio.com/docs/editor/versioncontrol). 
- **Live-sharing code**: Allows for collaborative code-editing like in Google Docs. Extremely useful for collaborating in the current work-from-home situation. Link [here](https://marketplace.visualstudio.com/items?itemName=MS-vsliveshare.vsliveshare-pack)
- **Markdown Previews**: I am currently writing this guide in Markdown using VSCode's native Markdown [preview functionalities](https://code.visualstudio.com/docs/languages/markdown). 
- **Editor Themes**: There are so many available themes on the VSCode marketplace to make the editor look and feel great. [Here](https://vscodethemes.com) is a website I used to find what I liked. 

## Absorption

### Description 

RadiativeTransfer.jl aims to revamp and modernize key atmospheric remote sensing tools. Specifically, it will enable the fast computation of atmospheric optical properties, full-polarized radiative transfer simulations, and commonly-used inversion routines.

By taking advantage of modern software tools, such as GPU acceleration and HPC computing, the software suite significantly accelerates computationally-intensive calculations and models, while keeping the interface easy-to-use for researchers and students.

The first module in this software suite is the Absorption.jl module. This module enables absorption cross-section calculations of atmospheric gases at different pressures, temperatures, and broadeners (Doppler, Lorentzian, Voigt). It uses the [HITRAN](https://hitran.org) energy transition database for calculations. While it enables lineshape calculations from scratch, it also allows users to create and save an interpolator object at specified wavelength, pressure, and temperature grids. It can perform these computations either on CPU or GPU. 

Please see installation instructions and more info on the repo's README [here](https://github.com/RupeshJey/RadiativeTransfer.jl). Further documentation is online [here](http://tofu.gps.caltech.edu:5055). 

### Demo

```julia
using RadiativeTransfer
using RadiativeTransfer.Absorption

## 
## STEP 1: Get the Hitran data
## 

# If you have a fixed-width file stored locally
hitran_data = read_hitran("path/to/file", mol=2, iso=1, ν_min=6000, ν_max=6400)

# If you would like to download the file from the hitran database
hitran_data = read_hitran(artifact"hitran_molec_id_2_CO2", mol=2, iso=1, ν_min=6000, ν_max=6400)

## 
## STEP 2: Create a model from parameters
## 

# These are some example models, but you can create/customize your model however you'd like. 
# Please make_hitran_model documentation for optional arguments
model_doppler = make_hitran_model(hitran_data, Doppler())
model_lorentz = make_hitran_model(hitran_data, Lorentz())
model_voigt_CPU = make_hitran_model(hitran_data, Voigt())
model_voigt_GPU = make_hitran_model(hitran_data, Voigt(), architecture=Architectures.GPU())

# If you would prefer to create an interpolation, and then interpolate the cross-section 
# at other wavelengths/pressures/temperatures, you can use make_interpolation_model as such:

ν_grid = 6000:0.01:6400
pressures = 250:250:1250
temperatures = 100:75:400

model_interp = make_interpolation_model(hitran_data, Voigt(), ν_grid, pressures, temperatures)

# Note: please see make_interpolation_model documentation for optional parameters

## 
## STEP 3: Calculate the absorption cross section with the created model
## 

# You can specify the wavelength grid, pressure, and temperature. 
absorption_cross_section(model_*, 6000:0.01:6400, 1000.1, 296.1)
```